/*
    qemu-start.c - Part of qemu-start, a package to start qemu nicely.
    Copyright (c) 2006  Frodo Looijaard <frodo@frodo.looijaard.name>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define _GNU_SOURCE /* asprintf */
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <net/if.h>
#include <linux/if_tun.h>

#define QEMU "/usr/bin/qemu"
#define USER "qemu"

/* Tiny code to open tap/tun device, and hand the fd to qemu.
   Run as root, drops to given user. */
int main(int argc, char *argv[])
{
	struct ifreq ifr;
	struct passwd *userdata;
	char *newargs[argc + 2];
	int i,fd,vlan,tapnr;
	char *ptr;
	char *command;
	char *display;
	char *xauthfile;

	/* Check parameters */
	if (argc < 4) {
		fprintf(stderr, "Usage: qemu-start TAPDEVNR VLAN SYSTEM <qemu options>...\n");
		fprintf(stderr, "  A device tapTAPDEVNR must have been configured.\n");
		fprintf(stderr, "  VLAN is the Virtual LAN number (just use 0 if unsure).\n");
		fprintf(stderr, "  SYSTEM is the system to emulate. Use \"\" for the current architecture\n");
		fprintf(stderr, "  (uses /usr/bin/qemm); else a binary /usr/bin/qemu-system-SYSTEM is used.\n");
		exit(1);
	}

	/* First parameter: TAPDEVNR */
	tapnr = strtol(argv[1],&ptr,0);
	if (*ptr) {
		fprintf(stderr, "Invalid value for TAPDEVNR parameter (use a number!)\n");
		exit(1);
	}

	/* Second parameter: VLAN */
	vlan = strtol(argv[2],&ptr,0);
	if (*ptr) {
		fprintf(stderr, "Invalid value for VLAN parameter (use a number!)\n");
		exit(1);
	}

	/* Open /dev/net/tun */
	fd = open("/dev/net/tun", O_RDWR);
	if (fd < 0) {
		perror("Could not open /dev/net/tun");
		exit(1);
	}

	/* Bind to the right tap device */
	memset(&ifr, 0, sizeof(ifr));
	ifr.ifr_flags = IFF_TAP | IFF_NO_PI;
	snprintf(ifr.ifr_name,IFNAMSIZ,"tap%d",tapnr);

	if (ioctl(fd, TUNSETIFF, (void *) &ifr) != 0) {
		perror("Could not get tap device");
		exit(1);
	}

	/* Fix X-display */
	userdata = getpwuid(getuid());
	if (! (xauthfile = getenv("XAUTHORITY")))
		asprintf(&xauthfile,"~%s/.Xauthority",userdata->pw_name);
	setuid(0);
	setgid(0);
	asprintf(&command,"xauth -f %s extract - $(echo $DISPLAY|sed \"s,^localhost,$(hostname)/unix,\") | xauth -f ~%s/.Xauthority merge -; chown %s: ~%s/.Xauthority",xauthfile,USER,USER,USER);
fprintf(stderr,"%s\n",command);
	system(command);

	/* Set correct userid. */
	userdata = getpwnam(USER);
	if (!userdata) {
		fprintf(stderr, "No user '%s'\n", USER);
		exit(1);
	}
	setgroups(0, NULL);
	setgid(userdata->pw_gid);
	if (setuid(userdata->pw_uid) != 0) {
		perror("setting uid");
		exit(1);
	}

	/* Clean the environment */
	display=getenv("DISPLAY");
	clearenv();
	setenv("HOME",userdata->pw_dir,1);
	setenv("DISPLAY",display,1);
	setenv("PATH","/usr/bin:/bin:/usr/local/bin",1);
	setenv("LOGNAME",userdata->pw_name,1);
	setenv("USER",userdata->pw_name,1);

	/* Create the right qemu invocation */
	if (!strlen(argv[3]))
		newargs[0] = QEMU;
	else
		asprintf(&newargs[0],"%s-system-%s",QEMU,argv[3]);
	newargs[1] = "-net";
	asprintf(&newargs[2],"nic,vlan=%d",vlan);
	newargs[3] = "-net";
	asprintf(&newargs[4],"tap,fd=%d,vlan=%d",fd,vlan);
	for (i = 4; i <= argc; i++)
		newargs[i+1] = argv[i];

	execvp(newargs[0], newargs);
	exit(1);
}
